/* 
	[990114] jono
*/

#include "version.h"
#include "alloc.h"
#include "exec.h"
#include "fail.h"
#include "gc.h"
#include "gc_ctrl.h"
#include "globals.h"
#include "intext.h"
#include "io.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "sys.h"
#include "interp.h"

// e_system.c
// 16May1997 e

// OSA AppleScript: compile, execute, display, error, dispose
// system() uses OSA AppleScript to execute its script arg

// see: IM: Interapplication Communication chapter 10

value mac_OSACompile( value v )
{
	#ifndef mjono	// [990114] jono
		
		failwith("mac_OSACompile not supported");
		return Val_unit;
		
	#else
		  OSAError err;
		  char *script;
		  AEDesc src = { typeNull, NULL };
		  long len;
		  value res;
		  OSAID obj = kOSANullScript;
		  Push_roots(r, 2);
		  r[1] = v;
		  r[0] = alloc( OSA_REC_SIZE, 0 );  
		  CHECK_INIT_OSA
		  len = strlen( (script = String_val( r[1] )) );
		  err = AECreateDesc( typeChar, script, len, &src );
		  if( err == noErr )
		    err = OSACompile( gScriptingComponent, &src, kOSAModeNull, &obj );
		  AEDisposeDesc( &src );
		done:
		  Field (r[0], OSA_REC_ERR) = Val_long( (short )err );
		  Field (r[0], OSA_REC_ID)  = copy_double( (double )obj );
		  res = r[0];
		  Pop_roots();
		  return res;
	#endif
}

value mac_OSAExecute( value v )
{
	#ifndef mjono	// [990114] jono
		
		failwith("mac_OSAExecute not supported");
		return Val_unit;
		
	#else
		  OSAError err;
		  OSAID rsv = kOSANullScript;
		  OSAID scp = (OSAID )Double_val( v );
		  value res;
		  Push_roots(r, 1);
		  r[0] = alloc( OSA_REC_SIZE, 0 );  
		  CHECK_INIT_OSA
		  err = OSAExecute( gScriptingComponent, scp, kOSANullScript, kOSAModeNull, &rsv );
		done:
		  Field (r[0], OSA_REC_ERR) = Val_long( (short )err );
		  Field (r[0], OSA_REC_ID)  = copy_double( (double )rsv );
		  res = r[0];
		  Pop_roots();
		  return res;
	#endif
}

value mac_OSADispose( value v )
{
	#ifndef mjono	// [990114] jono
		
		failwith("mac_OSADispose not supported");
		return Val_unit;
		
	#else
		  OSAError err;
		  OSAID rsv = (OSAID )Double_val( v );
		  CHECK_INIT_OSA
		  err = OSADispose( gScriptingComponent, rsv );
		done:
		  return Val_long( (short )err );;
	#endif
}

/*
static value copy_handle( Handle h )
{
  long len = GetHandleSize( h );
  value str = alloc_string( len );
  HLock( h );
  bcopy( *h, String_val( str ), len );
  HUnlock( h );
  return str;
}
*/

value mac_OSADisplay( value v ) // kOSAModeDisplayForHumans option?
{
	#ifndef mjono	// [990114] jono
		
		failwith("mac_OSADisplay not supported");
		return Val_unit;
		
	#else
		  OSAError err;
		  value res, str;
		  OSAID rsv = (OSAID )Double_val( v );
		  AEDesc dsc = { typeNull, NULL };
		  Push_roots(r, 1);
		  r[0] = alloc( OSA_REC_SIZE, 0 );  
		  CHECK_INIT_OSA
		  err = OSADisplay( gScriptingComponent, rsv, typeChar, kOSAModeNull, &dsc );
		  str = ( err == noErr ) ? copy_handle( dsc.dataHandle ) : copy_string( "" );
		  AEDisposeDesc( &dsc );
		done:
		  Field (r[0], OSA_REC_ERR) = Val_long( (short )err );
		  Field (r[0], OSA_REC_STR) = str;
		  res = r[0];
		  Pop_roots();
		  return res;

	#endif
}

value mac_OSAScriptError( value v )
{
	#ifndef mjono	// [990114] jono
		
		failwith("mac_OSAScriptError not supported");
		return Val_unit;
		
	#else
		  #pragma unused( v )
		  
		  OSAError err;
		  value res, str;
		  AEDesc dsc = { typeNull, NULL };
		  Push_roots(r, 1);
		  r[0] = alloc( OSA_REC_SIZE, 0 );  
		  CHECK_INIT_OSA
		  err = OSAScriptError( gScriptingComponent, kOSAErrorMessage, typeChar, &dsc );
		  str = ( err == noErr ) ? copy_handle( dsc.dataHandle ) : copy_string( "" );
		  AEDisposeDesc( &dsc );
		  err = OSAScriptError( gScriptingComponent, kOSAErrorNumber, typeShortInteger, &dsc );
		  if( err == noErr ) err = **(short **)dsc.dataHandle;
		  AEDisposeDesc( &dsc );
		done:
		  Field (r[0], OSA_REC_ERR) = Val_long( (short )err );
		  Field (r[0], OSA_REC_STR) = str;
		  res = r[0];
		  Pop_roots();
		  return res;
	#endif
}

// end.
